use codigocomentado::*;

use controllers::posts;
use dotenv::dotenv;
use misc::gen_title;
use tera::Tera;
use tokio::sync::OnceCell;

static TERA: OnceCell<Tera> = OnceCell::const_new();

async fn render<S: AsRef<str>>(name: S, context: &tera::Context) -> Result<String, ()> {
    let tera = TERA
        .get_or_init(|| async {
            let mut tera = Tera::default();
            for path in glob::glob("templates/gemini/*").unwrap().flatten() {
                let raw_path = path.clone();
                let filename = raw_path
                    .file_name()
                    .unwrap() // This should be safe (?
                    .to_str()
                    .unwrap() // ._.?
                    .split('.')
                    .next();
                tera.add_template_file(path, filename).unwrap()
            }
            tera
        })
        .await;
    tera.render(name.as_ref(), context).map_err(|_| ())
}

#[windmark::main]
async fn main() -> Result<(), Box<dyn std::error::Error>> {
    dotenv().ok();

    windmark::router::Router::new()
        .set_private_key_file("key.pem")
        .set_certificate_file("cert.pem")
        .set_languages(["es"])
        .set_port(1969)
        .enable_default_logger(true)
        .set_fix_path(true)
        .mount("/", move |_| {
            let (posts, _) = posts::get_posts(None);
            let mut context = tera::Context::new();

            context.insert("title", &gen_title());
            context.insert("posts", &posts);

            async move {
                windmark::response::Response::success(render("index", &context).await.unwrap())
            }
        })
        .mount("/post/:title", |request| {
            let title_splited: Vec<&str> = request
                .parameters
                .get("title")
                // This should be safe because its requires the parameter to enter to the route
                .unwrap()
                .split('-')
                .collect();
            let id = title_splited.last().unwrap().parse().unwrap_or(-1);
            async move {
                let result = match posts::get_post(id) {
                    Ok(post) => {
                        let mut context = tera::Context::new();
                        let content = post
                            .content
                            .replace("<pre>", "```")
                            .replace("</pre>", "```")
                            .replace("<b>", "")
                            .replace("</b>", "");
                        let content = md2gemtext::convert(&content);

                        context.insert("header", &gen_title());
                        context.insert("title", &post.title);
                        context.insert("created_at", &post.created_at);
                        context.insert("content", &content);
                        render("post", &context).await.unwrap()
                    }
                    Err(_) => "No se encontró ningún Post :c".into(),
                };
                windmark::response::Response::success(result)
            }
        })
        .set_error_handler(|_| {
            windmark::response::Response::permanent_failure("This route does not exist!")
        })
        .run()
        .await
}
